#!/usr/bin/python3
# -------------------------------------------------------------------------------

from pathlib import Path

import pytest

from pathtree.cli import PathtreeCLI, main
from pathtree.pathtree import PathTree

# -------------------------------------------------------------------------------


def find_and_print(tree, name, expected=True):

    node = tree.find(name)
    if node:
        if expected:
            print(node)
        else:
            raise FileExistsError(name)
    elif expected:
        raise FileNotFoundError(name)


# -------------------------------------------------------------------------------


class TestPathTree:
    def test_one(self):
        print()
        paths = [
            Path("/root"),
            Path("/root/a-f"),
            Path("/root/b-d"),
            Path("/root/b-d/g-f"),
            Path("/root/b-d/h-d"),
            Path("/root/b-d/h-d/m-f"),
            Path("/root/c-f"),
            Path("/root/d-d"),
            Path("/root/d-d/i-f"),
            Path("/root/d-d/j-d"),
            Path("/root/d-d/j-d/domain"),
            Path("/root/d-d/j-d/domain/kingdom"),
            Path("/root/d-d/j-d/domain/kingdom/phylum"),
            Path("/root/d-d/j-d/n-f"),
            Path("/root/d-d/j-d/red"),
            Path("/root/d-d/j-d/red/yellow"),
            Path("/root/d-d/j-d/red/yellow/blue"),
            Path("/root/e-f"),
            Path("/root/f-d"),
            Path("/root/f-d/k-f"),
            Path("/root/f-d/l-d"),
            Path("/root/f-d/l-d/m-f"),  # collision in basename
            Path("/root/f-d/l-d/o-f"),
        ]

        tree = PathTree()
        for path in paths:
            tree.addpath(path)

        tree.print()
        find_and_print(tree, "/root/d-d")
        find_and_print(tree, "/root/d-d/j-d")

    # -------------------------------------------------------------------------------

    def test_two(self):
        print()
        tree = PathTree()
        tree.addpath(Path("bob"))
        tree.addpath(Path("bob/george"))
        tree.addpath(Path("bob/sally"))
        tree.addpath(Path("bob/sally/ben"))
        tree.addpath(Path("bob/sally/ben/ralph"))

        tree.print()
        find_and_print(tree, "bob")

    # -------------------------------------------------------------------------------

    def test_three(self):
        print()
        tree = PathTree()
        tree.addpath(Path("top/bob"))
        tree.addpath(Path("top/bob/george"))
        tree.addpath(Path("top/bob/sally"))
        tree.addpath(Path("top/bob/sally/ben"))
        tree.addpath(Path("top/bob/sally/ben/ralph"))

        tree.print()
        find_and_print(tree, "top")
        find_and_print(tree, "top/bob")

    # -------------------------------------------------------------------------------

    @pytest.mark.skip(reason="collapse not implemented")
    def test_four(self):
        print()
        tree = PathTree()  # (collapse=True)
        tree.addpath(Path("top/bob"))
        tree.addpath(Path("top/bob/george"))
        tree.addpath(Path("top/bob/sally"))
        tree.addpath(Path("top/bob/sally/ben"))
        tree.addpath(Path("top/bob/sally/ben/ralph"))

        tree.print()
        find_and_print(tree, "top", expected=False)
        find_and_print(tree, "top/bob")

    # -------------------------------------------------------------------------------

    @pytest.mark.skip(reason="collapse not implemented")
    def test_four_2(self):
        print()
        tree = PathTree()  # (collapse=True)
        tree.addpath(Path("/top/bob"))
        tree.addpath(Path("/top/bob/george"))
        tree.addpath(Path("/top/bob/sally"))
        tree.addpath(Path("/top/bob/sally/ben"))
        tree.addpath(Path("/top/bob/sally/ben/ralph"))

        tree.print()
        find_and_print(tree, Path("/top"), expected=False)
        find_and_print(tree, Path("/top/bob"))

    # -------------------------------------------------------------------------------

    @pytest.mark.skip(reason="collapse not implemented")
    def test_five(self):
        print()
        tree = PathTree()  # (collapse=True)
        tree.addpath(Path("root/top/bob"))
        tree.addpath(Path("root/top/bob/george"))
        tree.addpath(Path("root/top/bob/sally"))
        tree.addpath(Path("root/top/bob/sally/ben"))
        tree.addpath(Path("root/top/bob/sally/ben/ralph"))

        tree.print()
        find_and_print(tree, Path("root"), expected=False)
        find_and_print(tree, Path("root/top"), expected=False)
        find_and_print(tree, Path("root/top/bob"))

    # -------------------------------------------------------------------------------

    @pytest.mark.skip(reason="collapse not implemented")
    def test_five_2(self):
        print()
        tree = PathTree()  # (collapse=True)
        tree.addpath(Path("/root/top/bob"))
        tree.addpath(Path("/root/top/bob/george"))
        tree.addpath(Path("/root/top/bob/sally"))
        tree.addpath(Path("/root/top/bob/sally/ben"))
        tree.addpath(Path("/root/top/bob/sally/ben/ralph"))

        tree.print()
        find_and_print(tree, Path("/root"), expected=False)
        find_and_print(tree, Path("/root/top"), expected=False)
        find_and_print(tree, Path("/root/top/bob"))

    # -------------------------------------------------------------------------------

    def test_prefix(self) -> None:
        print()
        PathtreeCLI.prefix = "woohoo"
        with pytest.raises(SystemExit) as err:
            main(["--help"])
        assert err.value.code == 0



"""An `os.PathLike` wrapper around `anytree.RenderTree`."""

import os
import textwrap
from pathlib import Path

import anytree


class PathTree:
    """An `os.PathLike` wrapper around `anytree.RenderTree`."""

    def __init__(self):
        """Create and return a new PathTree(object)."""

        self.rel_tree = None
        self.abs_tree = None
        self.tree = None

    def addpath(self, path):
        """Graft each node in path onto the tree.

        Unless it's already there, which is ok, like `mkdir -p`.

        Arguments:
            path:       can either be a string representing a pathname,
                        or an object implementing the os.PathLike interface
                        which returns a string, or another path object.
                        See https://docs.python.org/3/library/pathlib.html#pure-paths

        Returns:
            The anytree.Node(object) within self.tree for the given path, where:

            node.name       type(str) form of path.

            node._path      `path` as given; avoiding namespace collision with
                            `anytree.Node.path`.

        """

        # peek inside for string name
        if isinstance(path, Path):
            pathname = str(path)
        elif hasattr(path, "path") and isinstance(path.path, Path):
            pathname = str(path.path)
        else:
            pathname = str(path)
        # now treat it as opaque

        # already exist?
        if node := self.find(pathname):
            return node

        # no, create new node.
        node = anytree.Node(pathname, _path=path)

        if pathname == os.path.sep and not self.abs_tree:
            # just created "/".
            self.abs_tree = node
            return node

        # determine its parent
        parent = None
        dirname = os.path.dirname(str(path))

        if dirname:
            if dirname != os.path.sep and (parent := self.find(dirname)):
                pass
            else:
                parent = self.addpath(dirname)
        else:
            if not self.rel_tree:
                self.rel_tree = anytree.Node("<ROOT>")
            parent = self.rel_tree

        assert parent
        node.parent = parent
        return node

    def find(self, path):
        """Search for `path` in the tree.

        Arguments:
            path:       can either be a string representing a pathname,
                        or an object implementing the os.PathLike interface
                        which returns a string, or another path object.
                        See https://docs.python.org/3/library/pathlib.html#pure-paths

        Returns:        the Node found or None.
        """

        if isinstance(path, Path):
            pathname = str(path)
        elif hasattr(path, "path") and isinstance(path.path, Path):
            pathname = str(path.path)
        else:
            pathname = str(path)

        node = None
        if self.abs_tree:
            node = anytree.search.find(self.abs_tree, lambda node: node.name == pathname)

        if not node and self.rel_tree:
            node = anytree.search.find(self.rel_tree, lambda node: node.name == pathname)

        return node

    def render(
        self,
        *,
        style="round",
        indent=0,
        lengthen=0,
        dirname_short=True,
        dirname_wrap=0,
        basename_short=True,
        basename_wrap=0,
    ):
        """Docstring."""

        # pylint: disable=too-many-arguments

        for self.tree in (self.abs_tree, self.rel_tree):  # noqa: B020 Found for loop that...
            yield from self._render(
                style=style,
                indent=indent,
                lengthen=lengthen,
                dirname_short=dirname_short,
                dirname_wrap=dirname_wrap,
                basename_short=basename_short,
                basename_wrap=basename_wrap,
            )

    def _render(
        self,
        *,
        style="round",
        indent=0,
        lengthen=0,
        dirname_short=True,
        dirname_wrap=0,
        basename_short=True,
        basename_wrap=0,
    ):

        # pylint: disable=too-many-arguments
        # pylint: disable=too-many-branches
        # pylint: disable=too-many-locals

        if not self.tree:
            return

        if isinstance(style, str):
            if style == "ascii":
                style = anytree.AsciiStyle()
            elif style in ("DoubleStyle", "double"):
                style = anytree.DoubleStyle()
            elif style in ("ContRoundStyle", "round"):
                style = anytree.ContRoundStyle()
            elif style in ("ContRoundStyle", "square"):
                style = anytree.ContStyle()
            else:
                raise ValueError(f"invalid style {style!r}")

        indent = " " * indent if indent else ""

        if lengthen:
            style.vertical = self._lengthen(style.vertical, lengthen)
            style.cont = self._lengthen(style.cont, lengthen)
            style.end = self._lengthen(style.end, lengthen)

        #
        flag = False
        for pre, fill, node in anytree.RenderTree(self.tree, style=style):
            if node.name == "<ROOT>":
                flag = True
                continue
            if flag:
                pre = pre[4:]
                if fill:
                    fill = fill[4:]

            short = basename_short if node.is_leaf else dirname_short
            wrap = basename_wrap if node.is_leaf else dirname_wrap

            name = os.path.basename(node.name) if short else node.name

            # pylint: disable=protected-access
            if not wrap:
                yield indent + pre, name, node._path
            else:
                lines = textwrap.wrap(name, width=wrap)
                yield indent + pre, lines.pop(0), node._path
                for line in lines:
                    yield indent + fill, line, None

    @staticmethod
    def _lengthen(text, length):
        if not text or len(text) < 2:
            return text
        return text[:-2] + (text[-2] * length) + text[-2:]

    def print(self, *args, **kwargs):
        """Docstring."""

        for art, text, path in self.render(*args, **kwargs):
            print(art, text, path)



"""Command line interface."""

import argparse
import sys

from libcli import BaseCLI

from pathtree.pathtree import PathTree

__all__ = ["PathtreeCLI"]


class PathtreeCLI(BaseCLI):
    """Command line interface."""

    config = {
        # distribution name, not importable package name
        "dist-name": "rlane-pathtree",
    }

    # When used by another application, options are prefixed.
    # When used internally, options are not prefixed.
    prefix = None  # "pathtree"

    def init_parser(self) -> None:
        """Initialize argument parser."""

        self.ArgumentParser(
            prog=__package__,
            description=(
                "Read list of paths from standard input and print tree to standard output."
            ),
        )

    def add_arguments(self) -> None:
        """Add arguments to parser."""

        if not PathtreeCLI.prefix or len(PathtreeCLI.prefix) == 0:
            title_prefix = ""
            dest_prefix = ""
            prefix = "--"
        else:
            title_prefix = PathtreeCLI.prefix + " "
            dest_prefix = PathtreeCLI.prefix + "_"
            prefix = "--" + PathtreeCLI.prefix + "-"

        group = self.parser.add_argument_group(title_prefix + "formatting options")
        #
        arg = group.add_argument(
            prefix + "style",
            default="round",
            choices=["ascii", "double", "round", "square"],
            help="choose rendering style",
        )
        self.add_default_to_help(arg)

        arg = group.add_argument(
            prefix + "indent",
            default=0,
            type=int,
            metavar="INDENT",
            help="indent output with INDENT spaces",
        )
        self.add_default_to_help(arg)

        arg = group.add_argument(
            prefix + "lengthen",
            default=0,
            type=int,
            metavar="COLUMNS",
            help="lengthen horizontal lines by COLUMNS",
        )
        self.add_default_to_help(arg)

        #
        dirname_short = group.add_mutually_exclusive_group()
        dest = dest_prefix + "dirname_short"

        arg = dirname_short.add_argument(
            prefix + "dirname-short",
            action="store_true",
            dest=dest,
            help="with short dirnames",
        )
        self.add_default_to_help(arg)

        arg = dirname_short.add_argument(
            prefix + "dirname-long",
            action="store_false",
            dest=dest,
            help="with long dirnames",
        )
        self.add_default_to_help(arg)

        arg = dirname_short.add_argument(
            prefix + "dirname-wrap",
            default=66,
            type=int,
            metavar="COLUMN",
            help="wrap dirnames at COLUMN",
        )
        self.add_default_to_help(arg)

        #
        basename_short = group.add_mutually_exclusive_group()
        dest = dest_prefix + "basename_short"

        arg = basename_short.add_argument(
            prefix + "basename-short",
            action="store_true",
            dest=dest,
            help="with short basenames",
        )
        self.add_default_to_help(arg)

        arg = basename_short.add_argument(
            prefix + "basename-long",
            action="store_false",
            dest=dest,
            help="with long basenames",
        )
        self.add_default_to_help(arg)

        arg = basename_short.add_argument(
            prefix + "basename-wrap",
            default=66,
            type=int,
            metavar="COLUMN",
            help="wrap basenames at COLUMN",
        )
        self.add_default_to_help(arg)

        arg = self.parser.add_argument(
            "FILE",
            nargs="?",
            type=argparse.FileType(),
            default=sys.stdin,
            help="read paths from `FILE` instead of `stdin`",
        )

    def main(self) -> None:
        """Command line interface entry point (method)."""

        tree = PathTree()
        for line in self.options.FILE:
            tree.addpath(line.strip())

        for art, text, _ in tree.render(
            style=self.options.style,
            indent=self.options.indent,
            lengthen=self.options.lengthen,
            dirname_short=self.options.dirname_short,
            dirname_wrap=self.options.dirname_wrap,
            basename_short=self.options.basename_short,
            basename_wrap=self.options.basename_wrap,
        ):
            print(art + text)


def main(args: list[str] | None = None) -> None:
    """Command line interface entry point (function)."""
    PathtreeCLI(args).main()
